/* risc.cpp: (c)2015 CECS (TS, GL, RD) */

#include <iostream>

#include "risc.h"

/************** Added by Guantao Liu 08/25/2015 **************/
using namespace SageBuilder;
using namespace SageInterface;
using namespace std;

SgType* variableType = NULL;
SgExpression* PrimChnlVarRefExp = NULL;
SgExpression* SCChnlVarRefExp = NULL;
SgClassDefinition* PrimChnlClassDef = NULL;
SgClassDefinition* SCChnlClassDef = NULL;
SgClassDefinition* SCModuleClassDef = NULL;
std::map<SgNode*, int> FuncCallExptoID;
std::string OrigIncludePath;
std::string OrigLibraryPath;
std::string NewIncludePath;
std::string NewLibraryPath;

bool
is_derived_from(SgClassDefinition* derived_class,
                SgClassDefinition* base_class)
{
  SgBaseClassPtrList baseCls = derived_class->get_inheritances();

  // Iterate over all direct base classes
  for (Rose_STL_Container<SgBaseClass*>::iterator it = baseCls.begin(); 
       it != baseCls.end(); it++)
  {
    // check this base class
    if ((*it)->get_base_class()->search_for_symbol_from_symbol_table() ==
         base_class->get_declaration()->search_for_symbol_from_symbol_table())
    {
      return true;
    }

    // In case there exists no definition
    if (!(*it)->get_base_class()->get_definingDeclaration())
    {
      continue;
    }

    // get definition of base classes
    SgClassDefinition* recursive_def = isSgClassDeclaration((*it)->
      get_base_class()->get_definingDeclaration())->get_definition();

    // recursive search
    if (is_derived_from(recursive_def, base_class))
    {
      return true;
    }
  }

  return false;
}

class visitorTraversal: public AstSimpleProcessing
{
  public:
    visitorTraversal() {}
    virtual void visit(SgNode* astNode);
};

void
visitorTraversal::visit(SgNode* astNode)
{
  SgMemberFunctionDeclaration* memfuncdecl 
    = isSgMemberFunctionDeclaration(astNode);

  if (memfuncdecl != NULL)
  {
    // Instrument channel classes derived from ::sc_core::sc_prim_channel

    // if (risc::tools::is_derived_from_sc_prim_channel(memfuncdecl->
    // get_class_scope()))
    if (is_derived_from(memfuncdecl->get_class_scope(), PrimChnlClassDef))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_prim_channel.\n",
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration 
                = buildVariableDeclaration ("_sc_chnl_scoped_lock", variableType, 
                  buildCompoundInitializer(buildExprListExp(PrimChnlVarRefExp), 
                  PrimChnlVarRefExp->get_type()));

              prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument channel classes derived from ::sc_core::sc_channel

    // if (risc::tools::is_derived_from_sc_channel(memfuncdecl->
    // get_class_scope()))
    if (is_derived_from(memfuncdecl->get_class_scope(), SCChnlClassDef))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_channel.\n", 
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf 
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration 
                = buildVariableDeclaration ("_sc_chnl_scoped_lock", variableType, 
                  buildCompoundInitializer(buildExprListExp(SCChnlVarRefExp), 
                  SCChnlVarRefExp->get_type()));

              prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument module classes derived from ::sc_core::sc_module

    if (is_derived_from(memfuncdecl->get_class_scope(), SCModuleClassDef))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_module.\n",
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if (memfuncmodifier.isConstructor()) // constructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgExprStatement* assignStmt = buildAssignStatement(
                buildVarRefExp("m_instance_id", SCModuleClassDef),
                buildPlusPlusOp(buildVarRefExp("_OoO_Instance_Counter", 
                memfuncdecl->get_class_scope()), SgUnaryOp::postfix));

              prependStatement(assignStmt, block);
            }
          }          
        }
      }
    }
  }

  // Instrument Instance Counters
  SgClassDefinition* classdf = isSgClassDefinition(astNode);

  if (classdf != NULL)
  {
    if (is_derived_from(classdf, SCModuleClassDef))
    {
      // static unsigned int InstanceCounter
      SgVariableDeclaration* InstCntVarDecl
        = buildVariableDeclaration ("_OoO_Instance_Counter", buildUnsignedIntType());
      InstCntVarDecl->get_declarationModifier().get_storageModifier().setStatic();

      prependStatement(InstCntVarDecl, classdf);

      // initialize InstanceCounter
      SgVariableDeclaration* InstCntInitDecl
        = buildVariableDeclaration ("_OoO_Instance_Counter", buildUnsignedIntType(),
          buildAssignInitializer(buildIntVal(0)), classdf);

      insertStatementAfter(classdf->get_declaration(), InstCntInitDecl);
    }
  }

  // Instrument segment boundaries
  SgFunctionCallExp* funccallexp = isSgFunctionCallExp(astNode);
  if (funccallexp != NULL)
  {
    SgFunctionDeclaration* funcdecl = funccallexp->getAssociatedFunctionDeclaration();

    if (funcdecl != NULL)
    {
      if (funcdecl->get_name().getString() == "wait" ||
          funcdecl->get_name().getString() == "create_thread_process" ||
          funcdecl->get_name().getString() == "create_method_process" ||
          funcdecl->get_name().getString() == "create_cthread_process")
        if (FuncCallExptoID.find(funccallexp) != FuncCallExptoID.end())
          funccallexp->append_arg(buildIntVal(FuncCallExptoID[funccallexp]));
        else
          funccallexp->append_arg(buildIntVal(0));

      if (funcdecl->get_name().getString() == "create_thread_process" ||
          funcdecl->get_name().getString() == "create_method_process" ||
          funcdecl->get_name().getString() == "create_cthread_process")
        funccallexp->append_arg(buildVarRefExp("m_instance_id", SCModuleClassDef));
    }
  }
}

// Tailored backend for SystemC source files
// Modified from ~src/roseSupport/utility_functions.C

int risc_backend (SgProject* project, UnparseFormatHelp *unparseFormatHelp = NULL, UnparseDelegate* unparseDelagate = NULL);

int
risc_backend (SgProject* project, UnparseFormatHelp *unparseFormatHelp, UnparseDelegate* unparseDelagate)
{
  // DQ (7/12/2005): Introduce tracking of performance of ROSE.
  TimingPerformance timer ("AST Object Code Generation (backend):");

  int finalCombinedExitStatus = 0;

  if (project->get_binary_only() == true)
  {
    ROSE_ASSERT(project != NULL);

    // DQ (8/21/2008): Only output a message when we we use verbose option.
    if ( SgProject::get_verbose() >= 1 )
      printf ("Note: Binary executables are unparsed, but not passed to gcc as assembly source code \n");

    project->skipfinalCompileStep(true);
  }

  if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
    printf ("Inside of backend(SgProject*) \n");

  // printf ("   project->get_useBackendOnly() = %s \n",project->get_useBackendOnly() ? "true" : "false");
  if (project->get_useBackendOnly() == false)
  {
    // Add forward references for instantiated template functions and member functions 
    // (which are by default defined at the bottom of the file (but should be declared 
    // at the top once we know what instantiations should be built)).  They must be 
    // defined at the bottom since they could call other functions not yet declared in 
    // the file.  Note that this fixup is required since we have skipped the class template 
    // definitions which would contain the declarations that we are generating.  We might 
    // need that as a solution at some point if this fails to be sufficently robust.
    // if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL-2 )
    //      printf ("Calling fixupInstantiatedTemplates() \n");
    // DQ (9/6/2005): I think this is handled separately within post processing
    // (either that or they are just marked for output n the post processing)
    // fixupInstantiatedTemplates(project);

    // generate C++ source code
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->unparse() \n");

    project->unparse(unparseFormatHelp,unparseDelagate);

    /* Patch the static_cast errors (Guantao Liu, 07/29/2015) */
    char SysCmd[300];

//  std::cout << "\nStart patching the static_cast errors in the generated source files\n";
    std::cout << "RISC: patching ROSE limitations..."
            << std::endl;

    for (int i = 0; i < project->numberOfFiles(); i++)
    {
      sprintf(SysCmd, "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      std::cout << "  " << SysCmd << std::endl;

      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }

      sprintf(SysCmd, "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      std::cout << "  " << SysCmd << std::endl;

      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }

      sprintf(SysCmd, "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      std::cout << "  " << SysCmd << std::endl;

      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }
    }

    /* End */

    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      std::cout << "  " << "source file(s) generated" << std::endl;
  }

  std::cout << "RISC: compiling executable..."
            << std::endl;
  // DQ (1/25/2010): We have to now test for both numberOfFiles() and numberOfDirectories(),
  // or perhaps define a more simple function to use more directly.
  // if (project->numberOfFiles() > 0)
  if (project->numberOfFiles() > 0 || project->numberOfDirectories() > 0)
  {
    // Compile generated C++ source code with vendor compiler.
    // Generate object file (required for further template processing 
    // if templates exist).
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->compileOutput() \n");

      finalCombinedExitStatus = project->compileOutput();
  }
  else
  // if (project->get_compileOnly() == false)
  {
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("   project->get_compileOnly() = %s \n",project->get_compileOnly() ? "true" : "false");

    // DQ (5/20/2005): If we have not permitted templates to be instantiated during initial 
    // compilation then we have to do the prelink step (this is however still new and somewhat 
    // problematic (buggy?)).  It relies upon the EDG mechansisms which are not well understood.
    bool callTemplateInstantation = (project->get_template_instantiation_mode() == SgProject::e_none);

    if (callTemplateInstantation == true)
    {
      // DQ (9/6/2005): I think that this is no longer needed
      printf ("I don't think we need to call instantiateTemplates() any more! \n");
      ROSE_ASSERT(false);

      // The instantiation of templates can cause new projects (sets of source files) 
      // to be generated, but since the object files are already processed this is 
      // not an issue here.  A project might, additionally, keep track of the ASTs
      // associated with the different phases of instantions of templates.
      if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
        printf ("Calling instantiateTemplates() \n");

       printf ("Skipping template support in backend(SgProject*) \n");
       // instantiateTemplates (project);
    }

    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->link() \n");

    // DQ (10/15/2005): Trap out case of C programs where we want to make sure that we don't use the C++ compiler to do our linking!
    // This could be done in the 
    if (project->get_C_only() == true)
    {
      printf ("Link using the C language linker (when handling C programs) = %s \n","gcc");
      // finalCombinedExitStatus = project->link("gcc");
      finalCombinedExitStatus = project->link("gcc");
    }
    else
    {
      // Use the default name for C++ compiler (defined at configure time)
      if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
        printf ("Link using the default linker (when handling non-C programs) = %s \n","g++");
      finalCombinedExitStatus = project->link("g++");
    }

    // printf ("DONE with link! \n");
  }

  // Message from backend to user.
  // Avoid all I/O to stdout if useBackendOnly == true.
  // if (project->get_useBackendOnly() == false)
  if ( SgProject::get_verbose() >= 1 )
    std::cout << "source file(s) compiled with vendor compiler. (exit status = " << finalCombinedExitStatus << ").\n" << std::endl;

  // Set the final error code to be returned to the user.
  project->set_backendErrorCode(finalCombinedExitStatus);
  return project->get_backendErrorCode();
}

int
instrumentor (risc::Design& design, risc::sg::DataConflictTable* data_conflict_table, 
  risc::sg::EventConflictTable* event_conflict_table,
  risc::sg::TimeAdvancedTable* time_advance_table)
{
  SgProject* project = design.project_;
  ROSE_ASSERT (project != NULL);
  SgClassDeclaration* ChnlScpLockClassDecl = NULL;
  SgFunctionDeclaration* SCMainFuncDecl = NULL;


  // Find the class declarations of chnl_scoped_lock, sc_prim_channel, 
  // sc_channel and sc_module
  Rose_STL_Container<SgNode*> classDefinitionList 
    = NodeQuery::querySubTree(project, V_SgClassDefinition);
  for (Rose_STL_Container<SgNode*>::iterator it = classDefinitionList.begin(); 
       it != classDefinitionList.end(); it++)
  {
    SgClassDefinition* classdf = isSgClassDefinition(*it);
    if (classdf != NULL)
    {
      if (classdf->get_qualified_name().getString() == 
          "::sc_core::chnl_scoped_lock")
      {
        // SgClassDeclaration of "::sc_core::chnl_scoped_lock"
        ChnlScpLockClassDecl = classdf->get_declaration();  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_prim_channel")
      {
        // SgClassDefinition of "::sc_core::sc_prim_channel"
        PrimChnlClassDef = classdf;  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_channel")
      {
        // SgClassDefinition of "::sc_core::sc_channel"
        SCChnlClassDef = classdf;  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_module")
      {
        // SgClassDefinition of "::sc_core::sc_module"
        SCModuleClassDef = classdf;  
      }

      if (ChnlScpLockClassDecl && PrimChnlClassDef &&
          SCChnlClassDef && SCModuleClassDef)
        break;
    }
  }
  ROSE_ASSERT(ChnlScpLockClassDecl != NULL);
  ROSE_ASSERT(PrimChnlClassDef != NULL);
  ROSE_ASSERT(SCChnlClassDef != NULL);
  ROSE_ASSERT(SCModuleClassDef != NULL);

  variableType = ChnlScpLockClassDecl->get_type();
  ROSE_ASSERT(variableType != NULL);

  PrimChnlVarRefExp = buildVarRefExp("m_mutex", PrimChnlClassDef);
  ROSE_ASSERT(PrimChnlVarRefExp != NULL);

  SCChnlVarRefExp = buildVarRefExp("m_mutex", SCChnlClassDef);
  ROSE_ASSERT(SCChnlVarRefExp != NULL);

  // Find the function declaration of sc_main
  Rose_STL_Container<SgNode*> funcDeclarationList 
    = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);
  for (Rose_STL_Container<SgNode*>::iterator it = funcDeclarationList.begin(); 
       it != funcDeclarationList.end(); it++)
  {
    SgFunctionDeclaration* funcdecl = isSgFunctionDeclaration(*it);
    if (funcdecl != NULL)
    {
      if (funcdecl->get_name().getString() == "sc_main"
        && funcdecl->get_definition() != NULL)
      {
        // SgFunctionDeclaration of "sc_main"
        SCMainFuncDecl = funcdecl;  
        break;
      }
    }
  }
  ROSE_ASSERT(SCMainFuncDecl != NULL);

  // Build the traversal object
  visitorTraversal exampleTraversal;

  // Call the traversal starting at the project node of the AST
  exampleTraversal.traverseInputFiles(project, preorder);

  // Output the data conflict table
  int DataConflTableSize = data_conflict_table->size_of_conflict_table_;
  std::vector<SgExpression *> DataConflTableValue;
  for (int i = 0; i < DataConflTableSize; i++)
    for (int j = 0; j < DataConflTableSize; j++)
      DataConflTableValue.push_back(buildIntVal(data_conflict_table->conflict_table_[i][j]));

  SgVariableDeclaration* DataConflTableDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Data_Conflict_Table",
      buildArrayType(buildBoolType(),
      buildIntVal(DataConflTableSize * DataConflTableSize)), 
      buildAggregateInitializer(buildExprListExp(DataConflTableValue)));

  insertStatementBefore(SCMainFuncDecl, DataConflTableDecl);

  // Output the data conflict table size
  SgVariableDeclaration* DataConflTableSizeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Data_Conflict_Table_Size",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(DataConflTableSize)));

  insertStatementBefore(SCMainFuncDecl, DataConflTableSizeDecl);

  // Output the event notification table
  int EventNotifyTableSize = event_conflict_table->size_of_conflict_table_;
  std::vector<SgExpression *> EventNotifyTableValue;
  for (int i = 0; i < EventNotifyTableSize; i++)
    for (int j = 0; j < EventNotifyTableSize; j++)
      EventNotifyTableValue.push_back(buildIntVal(event_conflict_table->conflict_table_[i][j]));

  SgVariableDeclaration* EventNotifyTableDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Event_Notify_Table",
      buildArrayType(buildBoolType(), 
      buildIntVal(EventNotifyTableSize * EventNotifyTableSize)), 
      buildAggregateInitializer(buildExprListExp(EventNotifyTableValue)));

  insertStatementBefore(SCMainFuncDecl, EventNotifyTableDecl);

  // Output the event notification table size
  SgVariableDeclaration* EventNotifyTableSizeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Event_Notify_Table_Size",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(EventNotifyTableSize)));

  insertStatementBefore(SCMainFuncDecl, EventNotifyTableSizeDecl);

  // Output the segment and instance id to index lookup table for conflict tables
  int MaxNumofInst = data_conflict_table->max_instances_;
  int NumofSeg = data_conflict_table->number_of_segments_;
  std::vector<SgExpression *> ConflictLookupTableValue;
  for (int i = 0; i < MaxNumofInst + 1; i++)
    for (int j = 0; j < NumofSeg; j++)
    {
      //ConflictLookupTableValue.push_back(buildIntVal(
        //data_conflict_table->instance_id_to_index_lookup_[i][j]));

      // Temporal patch: all unanalyzed segment boundaries are given ID 0 TODO
      if (j == 0 && data_conflict_table->instance_id_to_index_lookup_[i][j] == -1)
        ConflictLookupTableValue.push_back(buildIntVal(0));
      else
        ConflictLookupTableValue.push_back(buildIntVal(
          data_conflict_table->instance_id_to_index_lookup_[i][j]));
    }

  SgVariableDeclaration* ConflictLookupTableDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Conflict_Index_Lookup_Table",
      buildArrayType(buildIntType(),
      buildIntVal((MaxNumofInst + 1) * NumofSeg)), 
      buildAggregateInitializer(buildExprListExp(ConflictLookupTableValue)));

  insertStatementBefore(SCMainFuncDecl, ConflictLookupTableDecl);

  // Output the maximum number of module instances and number of segments
  SgVariableDeclaration* MaxNumofInstDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Max_Number_of_Instances",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(MaxNumofInst)));

  insertStatementBefore(SCMainFuncDecl, MaxNumofInstDecl);

  SgVariableDeclaration* NumofSegDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Number_of_Segments",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(NumofSeg)));

  insertStatementBefore(SCMainFuncDecl, NumofSegDecl);

  // Output the current time advance table
  int CurrTimeAdvTableSize = time_advance_table->size_;

  std::vector<SgExpression *> CurrTimeAdvTableTimeValue;
  for (int i = 0; i < CurrTimeAdvTableSize; i++)
  {
    if (time_advance_table->current_time_advance_[i].infinite_)
      CurrTimeAdvTableTimeValue.push_back(buildLongLongIntVal(-1)); // -1 stands for infinity
    else
      CurrTimeAdvTableTimeValue.push_back(buildLongLongIntVal(
        time_advance_table->current_time_advance_[i].time_units_));
  }

  SgVariableDeclaration* CurrTimeAdvTableTimeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Time",
      buildArrayType(buildLongLongType(), buildIntVal(CurrTimeAdvTableSize)), 
      buildAggregateInitializer(buildExprListExp(CurrTimeAdvTableTimeValue)));

  std::vector<SgExpression *> CurrTimeAdvTableDeltaValue;
  for (int i = 0; i < CurrTimeAdvTableSize; i++)
  {
    if (time_advance_table->current_time_advance_[i].infinite_)
      CurrTimeAdvTableDeltaValue.push_back(buildIntVal(-1)); // -1 stands for infinity
    else
      CurrTimeAdvTableDeltaValue.push_back(buildIntVal(
        time_advance_table->current_time_advance_[i].delta_count_));
  }

  SgVariableDeclaration* CurrTimeAdvTableDeltaDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Delta",
      buildArrayType(buildIntType(), buildIntVal(CurrTimeAdvTableSize)), 
      buildAggregateInitializer(buildExprListExp(CurrTimeAdvTableDeltaValue)));

  insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableTimeDecl);
  insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableDeltaDecl);

  // Output the current time advance table size
  SgVariableDeclaration* CurrTimeAdvTableSizeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Size",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(CurrTimeAdvTableSize)));

  insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableSizeDecl);

  // Output the next time advance table
  int NextTimeAdvTableSize = time_advance_table->size_;

  std::vector<SgExpression *> NextTimeAdvTableTimeValue;
  for (int i = 0; i < NextTimeAdvTableSize; i++)
  {
    if (time_advance_table->next_time_advance_[i].infinite_)
      NextTimeAdvTableTimeValue.push_back(buildLongLongIntVal(-1)); // -1 stands for infinity
    else
      NextTimeAdvTableTimeValue.push_back(buildLongLongIntVal(
        time_advance_table->next_time_advance_[i].time_units_));
  }

  SgVariableDeclaration* NextTimeAdvTableTimeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Time",
      buildArrayType(buildLongLongType(), buildIntVal(NextTimeAdvTableSize)), 
      buildAggregateInitializer(buildExprListExp(NextTimeAdvTableTimeValue)));

  std::vector<SgExpression *> NextTimeAdvTableDeltaValue;
  for (int i = 0; i < NextTimeAdvTableSize; i++)
  {
    if (time_advance_table->next_time_advance_[i].infinite_)
      NextTimeAdvTableDeltaValue.push_back(buildIntVal(-1)); // -1 stands for infinity
    else
      NextTimeAdvTableDeltaValue.push_back(buildIntVal(
        time_advance_table->next_time_advance_[i].delta_count_));
  }

  SgVariableDeclaration* NextTimeAdvTableDeltaDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Delta",
      buildArrayType(buildIntType(), buildIntVal(NextTimeAdvTableSize)), 
      buildAggregateInitializer(buildExprListExp(NextTimeAdvTableDeltaValue)));

  insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableTimeDecl);
  insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableDeltaDecl);

  // Output the next time advance table size
  SgVariableDeclaration* NextTimeAdvTableSizeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Size",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(NextTimeAdvTableSize)));

  insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableSizeDecl);

  // Output the segment id to index lookup table for time advance tables
  int TimeAdvLookupTableSize = time_advance_table->size_;
  std::vector<SgExpression *> TimeAdvLookupTableValue;
  for (int i = 0; i < TimeAdvLookupTableSize; i++)
      TimeAdvLookupTableValue.push_back(buildIntVal(time_advance_table->get_segment_id(i)));

  SgVariableDeclaration* TimeAdvLookupTableDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Time_Advance_Index_Lookup_Table",
      buildArrayType(buildIntType(), buildIntVal(TimeAdvLookupTableSize)), 
      buildAggregateInitializer(buildExprListExp(TimeAdvLookupTableValue)));

  insertStatementBefore(SCMainFuncDecl, TimeAdvLookupTableDecl);

  // Output the size of the segment id to index lookup table
  SgVariableDeclaration* TimeAdvLookupTableSizeDecl 
    = buildVariableDeclaration ("sc_core::_OoO_Time_Advance_Index_Lookup_Table_Size",
      buildUnsignedIntType(), 
      buildAssignInitializer(buildIntVal(TimeAdvLookupTableSize)));

  insertStatementBefore(SCMainFuncDecl, TimeAdvLookupTableSizeDecl);

  // Test the new AST
  AstTests::runAllTests(project);

  // Change the command line argument list
  Rose_STL_Container<std::string> argList = project->get_originalCommandLineArgumentList();
  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    if (it->compare(OrigIncludePath) == 0)
      *it = NewIncludePath;

    if (it->compare(OrigLibraryPath) == 0)
      *it = NewLibraryPath;
  }
  project->set_originalCommandLineArgumentList(argList);

  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    Rose_STL_Container<std::string> argListFile = file.get_originalCommandLineArgumentList();
    for (Rose_STL_Container<std::string>::iterator it = argListFile.begin();
         it != argListFile.end(); it++)
    {
      if (it->compare(OrigIncludePath) == 0)
        *it = NewIncludePath;

      if (it->compare(OrigLibraryPath) == 0)
        *it = NewLibraryPath;
    }
    file.set_originalCommandLineArgumentList(argListFile);
  }

  return risc_backend(project);
}

void
ParseCommandlineOptions(int &argc, char** &argv)
{
  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    // -risc:origincl:
    pos = it->find("-risc:origincl:");
    if (pos != string::npos)
    {
      it->erase(pos + 1, 14);
      OrigIncludePath = *it;
    }

    // -risc:origlib:
    pos = it->find("-risc:origlib:");
    if (pos != string::npos)
    {
      it->erase(pos + 1, 13);
      OrigLibraryPath = *it;
    }

    // -risc:newincl:
    pos = it->find("-risc:newincl:");
    if (pos != string::npos)
    {
      it->erase(pos + 1, 13);
      NewIncludePath = *it;
      RemovedIter.push_back(it);
    }

    // -risc:newlib:
    pos = it->find("-risc:newlib:");
    if (pos != string::npos)
    {
      it->erase(pos + 1, 12);
      NewLibraryPath = *it;
      RemovedIter.push_back(it);
    }
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);
}

/**************** End  Guantao Liu 08/25/2015 ****************/

int main(int argc, char* argv[])
{
/************** Added by Guantao Liu 09/04/2015 **************/
  ParseCommandlineOptions(argc, argv);
/**************** End  Guantao Liu 09/04/2015 ****************/

  std::cout //<< std::endl
//            << "start processing file: " __FILE__
//            << " ..."
            << "RISC"
            << " starting..."
            << std::endl;

  std::cout << "RISC" << ": parsing, building ROSE IR..."
            << std::endl;
  risc::Design d(argc,argv);

  std::cout << "RISC" << ": analyzing, building RISC SG..."
            << std::endl;
  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      d.systemc_definitions_->sc_wait_funcs_);
  risc::sg::SegmentGraph sg(function_call_boundary);

  std::cout << "RISC" << ": listing model hierarchy..."
            << std::endl;
  for(risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) {

    std::cout << "  Module: " <<  (*iter_module)->get_name() << std::endl;

    for(risc::ThreadVectorIter
        iter_thread = (*iter_module)->threads_.begin();
        iter_thread != (*iter_module)->threads_.end();
        iter_thread++) {

      std::cout << "    " << (*iter_thread)->get_name() << std::endl;

      sg.add_function(*iter_thread);
/************** Added by Guantao Liu 08/28/2015 **************/
      FuncCallExptoID[(*iter_thread)->starting_node_] = (*iter_thread)->starting_id_;
/**************** End  Guantao Liu 08/28/2015 ****************/
    }

//    std::cout << std::endl;
  }

  sg.clean_graph();

  std::cout << "RISC" << ": analyzing, building conflict tables..."
            << std::endl;
  // Time advanced tables
  risc::sg::TimeAdvancedTable time_advanced_table(sg);
  time_advanced_table.print_current_time_advanced_table(
                          "current_time_advanced_table.html");
  time_advanced_table.print_next_time_advanced_table(
                          "next_time_advanced_table.html");

  // Event conflict table
  risc::sg::EventConflictTable
    event_conflict_table(sg, d.path_instance_mapper_, true);
  event_conflict_table.print_conflict_table("event_conflict_table.html");
  event_conflict_table.print_lookup_table("conflict_lookup_table.html");

  // Data conflict table
  risc::sg::DataConflictTable
    data_conflict_table(sg, d.path_instance_mapper_, true);
  data_conflict_table.print_conflict_table("data_conflict_table.html");

//  std::cout << "end processing file: " __FILE__
//            << " ..."
//            << std::endl
//            << std::endl;

/************** Added by Guantao Liu 08/24/2015 **************/
  std::cout << "RISC" << ": generating SG visualization..."
            << std::endl;
  sg.print_graph("segment_graph.dot");

  std::cout << "RISC" << ": instrumenting, generating parallel model..."
            << std::endl;
  std::pair<risc::sg::VertexIterator, risc::sg::VertexIterator> vertices_range = vertices(sg.graph_);
  for (risc::sg::VertexIterator iter = vertices_range.first; iter != vertices_range.second; iter++)
  {
    const SgExprStatement* waitexprstmt = isSgExprStatement(sg.graph_[*iter].boundary_node_);
    if (waitexprstmt != NULL)
    {
      SgFunctionCallExp* funccallexp = isSgFunctionCallExp(waitexprstmt->get_expression());
      if (funccallexp != NULL)
        FuncCallExptoID[funccallexp] = sg.graph_[*iter].id_;
    }
  }

  instrumentor(d, &data_conflict_table, &event_conflict_table, &time_advanced_table);
/**************** End  Guantao Liu 08/24/2015 ****************/

  std::cout << std::endl
            << "RISC"
            << " done."
            << std::endl;

  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
