#ifndef SEGMENT_H_INCLUDED_
#define SEGMENT_H_INCLUDED_

#include<string>
#include<vector>
#include<set>
#include<iostream>

#include "rose.h"

namespace risc {

namespace sg {

class Segment {

public:
  //typedef std::vector<const SgStatement*> Statements;
  typedef std::vector<SgNode*> Expressions;
  boost::default_color_type color;

  enum Status {Untouched = 0, Touched = 1};
  Status color_for_communication_graph_;
  Status helper_for_graph_algorithms_;

  Expressions expressions_;

  bool debugging_;
  std::set<SgNode*> conflicts_;

  Segment(SgNode const * const boundary_node);

  Segment(const Segment &other);

  /**
   * \function analyze_variable_accesses
   * \brief    This function determines which variable are read and written
   *           in this segment
   */
  void analyze_segment();

  /**
   * \function analyze_expression
   * \brief    This function analyzes recursively all read and write accesses
   *           in this expression
   */
  void analyze_expression(SgNode *expression, bool left_of_assignment,
    bool read_write, bool member_access);

  friend std::ostream& operator<< (std::ostream &out, const Segment &segment)
  {
    return out;
  }

  /**
   * \function add_expression
   * \brief    This function add the given node to the vector of expressions
   */
  void add_expression(SgNode* node);

  /**
   * \function operator==
   * \brief    If two segments are equal the return value is true
   */
  bool operator==(const Segment &other);

  /*
   * \brief Merges the other segment in this segment.
   *        No statement will be doubled
   */
  void merge(const Segment &other);

  /**
   * \brief This is the unique id of the segment
   */
  int id_;

  /**
   * \brief This is the associated boundary of the segment
   */
  SgNode const * const boundary_node_;

  /**
   * \brief This static counter will be used to generate id's for the
   *        individual segments.
   */
  static int id_counter;

  /**
   * \brief This color is used for some boost algorithms
   */
  boost::default_color_type color_;

  /**
   * \brief Set of all SgVariableSymbol which are read in this segment
   */
  std::set<SgVariableSymbol*> read_variables_;

  /**
   * \brief Set of all SgVariableSymbol which are written in this segment
   */
  std::set<SgVariableSymbol*> write_variables_;

  /**
   * \brief If there is any pointer or reference operations, this flag will be
   *        set
   */
  bool in_conflict_with_all_other_segments;

//  /**
//   * \brief 
//   */
//  std::set<SgVariableSymbol*> read_references_;
//
//  std::set<SgVariableSymbol*> write_references_;
};



} // end of namespace sg

} // end of namespace risc

#endif /* SEGMENT_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
