#ifndef CTHREAD_ANALYZER_H_INCLUDED_
#define CTHREAD_ANALYZER_H_INCLUDED_

#include "rose.h"
#include "../internal_representation/cthread.h"
#include "../internal_representation/module.h"

#include <vector>
#include <set>
#include <utility>

namespace risc
{

namespace sa
{
/// CThread -> var, var, var ...
typedef std::pair<CThread*, std::set<SgVariableSymbol*> >
            CThreadUninitializedVariables;
typedef std::set<CThreadUninitializedVariables>
            CThreadUninitializedVariablesSet;

/// ResetSignal -> cthread, cthread, cthread ...
typedef std::pair<ResetSignal*, CThreadVector> ResetSignalSensitiveCThreads;
typedef std::set<ResetSignalSensitiveCThreads> ResetSignalSensitiveCThreadsSet;

/// ResetSignal -> var, var, var ...
typedef std::pair<ResetSignal*, std::set<SgVariableSymbol*> >
            ResetSignalUninitializedVariables;
typedef std::set<ResetSignalUninitializedVariables>
            ResetSignalUninitializedVariablesSet;


typedef std::pair<CThread*, ResetSignal*>
            CThreadResetSignalPair;
typedef std::pair<CThreadResetSignalPair, std::vector<SgExpression*> >
            CThreadResetSignalExpressions;
typedef std::vector<CThreadResetSignalExpressions>
            CThreadResetSignalExpressionsVector;

/**
 * \brief This function analyzes a given function and returns the first
 *        occurrence of a wait call
 */
SgFunctionCallExp* find_first_wait_stmt(SgFunctionDefinition *func_def);


/**
 * \brief This function searchs for all function calls which are done
 *        before the first wait call statement. The search doesnt respect
 *        function calls in loops, if statements, or other nested constructs.
 * \return A set of function calls
 */
std::set<SgFunctionCallExp*>
traverse_for_non_nested_func_calls(SgFunctionDefinition *func_def);


/**
 * \brief This function collects all non nested expressions before and after
 *        the first wait stmt call
 */
void
collect_non_nested_expressions(SgFunctionDefinition *func_def,
    std::vector<SgExpression*> &before_wait,
    std::vector<SgExpression*> &after_wait);

CThreadUninitializedVariablesSet
most_pessimistic_analysis(risc::Module &module);

CThreadUninitializedVariablesSet
inter_thread_analysis(Module &module, CThreadVector *subset_of_cthreads = NULL);

ResetSignalUninitializedVariablesSet
inter_reset_cthread_analysis(Module &module);

ResetSignalUninitializedVariablesSet
inter_async_reset_cthread_analysis(Module &module);

CThreadResetSignalExpressionsVector
get_cthread_reset_signal_expressions(Module &module);

CThreadResetSignalExpressionsVector
get_cthread_async_reset_signal_expressions(Module &module);

void
member_variable_filter(risc::Module &module, std::vector<SgNode*> variable);

void
get_recursivly_nested_expressions_in_function(
    SgFunctionDefinition *root_function,
    std::vector<SgExpression*> &expressions);

std::set<SgFunctionDefinition*>
get_called_functions_and_expressions(SgFunctionDefinition *root_function,
    std::vector<SgExpression*> &expressions);

} // end namesapce sa

} // end namespace risc

#endif /* CTHREAD_ANALYZER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
