#include "class_is_derived_from.h"

#include "../internal_representation/design.h"

bool
risc::tools::is_derived_from(SgClassDefinition* derived_class,
    SgClassDefinition* base_class)
{
  // FIXME This is a work around for the broken AST link of the base
  // class of sc_event_and_list
  if(base_class == NULL) {
    return false;
  }

  SgBaseClassPtrList base_classes = derived_class->get_inheritances();
  SgBaseClassPtrList::iterator iter = base_classes.begin();

  // iterate all direct base classes
  for(;iter != base_classes.end();iter++) {

    // check current base class
    if((*iter)->get_base_class()->search_for_symbol_from_symbol_table() ==
      base_class->get_declaration()->search_for_symbol_from_symbol_table()) {

      return true;
    }

    // In this case there exists no definition
    if(!(*iter)->get_base_class()->get_definingDeclaration()) {
      std::cout << "Cannot find any definition of class: "
                << SageInterface::get_name((*iter)->get_base_class())
                << std::endl;
      continue;
    }

    // get definition of base classes
    SgClassDefinition* recursive_def =
      isSgClassDeclaration((*iter)->get_base_class()->
          get_definingDeclaration())->get_definition();

    // recursive search
    if(is_derived_from(recursive_def, base_class)) {
      return true;
    }
  }

  return false;
}


bool
risc::tools::is_derived_from_sc_module(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             risc::get_design()->systemc_definitions_->sc_module_);
}

bool
risc::tools::is_derived_from_sc_port(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
              risc::get_design()->systemc_definitions_->sc_port_);
}

bool
risc::tools::is_derived_from_sc_port_base(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             risc::get_design()->systemc_definitions_->sc_port_base_);
}

bool
risc::tools::is_derived_from_sc_port_base(SgVarRefExp* var_ref)
{
  SgClassDefinition *class_def =
    isSgClassDeclaration(
        isSgClassDeclaration(
            isSgClassDeclaration(
              isSgClassType(var_ref->get_type())->get_declaration()))->
        get_definingDeclaration())->
    get_definition();

  return is_derived_from(class_def,
             risc::get_design()->systemc_definitions_->sc_port_base_);
}

bool
risc::tools::is_derived_from_sc_prim_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             risc::get_design()->systemc_definitions_->sc_prim_channel_);
}

bool
risc::tools::is_derived_from_sc_channel(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class,
             risc::get_design()->systemc_definitions_->sc_channel_);
}

bool
risc::tools::is_sc_in_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_in_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_out_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_port(SgTemplateInstantiationDefn* template_class)
{
  SgClassDeclaration *class_decl = template_class->get_declaration();
  SgTemplateInstantiationDecl *templ_class_decl =
    isSgTemplateInstantiationDecl(class_decl);

  return templ_class_decl->get_templateDeclaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_inout_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_in_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_in_clk_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_out_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_out_clk_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_inout_clk(SgClassDefinition *definition)
{
  return definition->get_declaration()->get_type() ==
    risc::get_design()->systemc_definitions_->sc_inout_clk_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_event(SgClassDefinition* derived_class)
{
  return is_derived_from(derived_class, risc::get_design()->
             systemc_definitions_->sc_event_)
    || (derived_class->get_declaration()->get_type() ==
         risc::get_design()->systemc_definitions_->sc_event_->
             get_declaration()->get_type());
}

bool
risc::tools::is_sc_event(SgVarRefExp* var_ref)
{
  return var_ref->get_type() ==
    risc::get_design()->systemc_definitions_->sc_event_->get_declaration()->
        get_type();
}

bool
risc::tools::is_sc_in_bool(SgClassDefinition* definition)
{
  SgTemplateInstantiationDefn *temp_inst_def =
    isSgTemplateInstantiationDefn(definition);

  if(temp_inst_def) {

    if(is_sc_in_port(temp_inst_def))  {

      SgTemplateInstantiationDecl *temp_inst_decl =
        isSgTemplateInstantiationDecl(temp_inst_def->get_declaration());

      if(temp_inst_decl->get_templateArguments().size() ==1 &&
          isSgTypeBool((temp_inst_decl->get_templateArguments().at(0))->
              get_type())) {
        return true;
      }
    }
  }

  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
